﻿' ============================================================
' File Name:  Demo_360X.vb
' Author(s):  Kelly Spiller, Based Config360X by Jim Lamberson 
' Sensoray Inc. 2021
' Desription: Demo Application for 3604-2 CPS Power Supply
'             usage of the c3600 middleware API
' ============================================================
' History
' Ver.1.0.0 - Initial Release
' Ver.1.1.0 - Added Logging for V and I meter's on all devices
' Ver.1.2.0 -
'  - Fixed Scrollbar settings for current  setpoint.
'  - Initialize V/I/T average to current meter reading.
'  - Fixed out of range when setting corona ProgressBar.
'  - Fixed Reading and display of initial/current Filament setpoint.
' Ver.1.3.0 -
'  - Remove Device combo-box.
'  - Dynamically re-size application if 1 or 2 clusters found. 
'    (Support use of single cluster systems like the 2600 power supply)
' Ver.2.0.0 -
'  - Dynamically re-size application and add controls as needed for 
'  - suplies with any number of clusters.
'  - Added progress bar to indicate progress of opening power supplies 
'  - with a large number of devices.
'    (Uses c3600.dll ver 4.1.0)
' Ver.2.1.0 - 
'  - Renamed project Demo_360X (from Demo_3604_2)
'    (Uses c3600.dll ver 4.1.0)
' Ver.2.2.0 - 
'  - Updated c3600.vb. Adds devpolarity property to DEV_INFO and DEV_CONFIG Structures.
'  - Added DLL version checking to OpenSystem()
'    (Uses c3600.dll ver 4.2.0)
'
' Update VERSON in:
'    ...\My Project\AssemblyInfo.vb
'    APP_VERSION_xx below
' 
' ============================================================






' Print and Display Constants defined by Microsoft Visual Basic Namespace: Microsoft.VisualBasic
' vbCrLf,vbCr,vbLf,vbNewLine (same as vbCrLf),vbNullChar,vbNullString,vbObjectError
' vbTab,vbBack (Chr-8),vbFormFeed (Chr-12),vbVerticalTab (Chr-11)
'

'Imports System
Imports System.IO ' Needed for File

' Combine each device in the power supply with the Windows controls associated with it.
' This allows generic handler routines that can figure out which device to control 
' based on which control was triggered
Structure DEVLIST
    Public valid As Integer                        ' Indicates C3600_GetDevConfig(sysid, i,... ) is valid for ith entry in sysdev_list
    'Public chksum As Int32                        ' Valid configuration checksum
    Public dev As DEV_CONFIG                       ' Device returned by GetDevConfig(sysid, i,... )
    '
    Public Grp As System.Windows.Forms.GroupBox    ' Group to hold controls
    ' Static controls
    Public ChkValidDevConfig As System.Windows.Forms.CheckBox  ' ChkValidDevConfig_A_0...
    Public DscrSetpoint      As System.Windows.Forms.Label ' DscrSetpoint_A_0...   static
    Public DscrVMeter        As System.Windows.Forms.Label ' DscrVMeter_A_0...     static
    Public DscrTMeter        As System.Windows.Forms.Label ' DscrTMeter_A_0...     static
    Public DscrIMeter        As System.Windows.Forms.Label ' DscrIMeter_A_0...     static
    Public LblIMeterUnits    As System.Windows.Forms.Label ' LblIMeterUnits_A_0... static / modified in OpenSystem() when discovered
    ' Control inputs
    Public sb_setpoint As Windows.Forms.HScrollBar ' scrollSetpoint_A_0, A_1, B_0, B_1, F_0, F_1
    Public cb_outen As Windows.Forms.CheckBox      ' chkOutputEnable_A_0, ...
    Public cb_meteren As Windows.Forms.CheckBox    ' chkMeterScan_A_0, ...
    ' Control outputs
    Public LblSetpoint As Windows.Forms.Label      ' lblSetpoint_A_0, ...
    Public LblVMeter As Windows.Forms.Label        ' LblVMeter_A_0, ...
    Public LblIMeter As Windows.Forms.Label        ' LblIMeter_A_0, ...
    Public LblTMeter As Windows.Forms.Label        ' LblTMeter_A_0, ...
    Public avgVMeter As Double                     ' Average Voltage
    Public avgIMeter As Double                     ' Average Current
    Public avgTMeter As Double                     ' Average Temperature
    Public lasti As Double   ' Last volt or amp reading for the Accelerator Corona calculation
    Public lastdi1 As Double ' Last difference between curent and previous volt/current reading
    Public det As Double     ' Last corona level calculated

    Public cb_Log As Windows.Forms.CheckBox        ' cb_Log_A_0, ...
    Public EnableLog As Boolean
    Public TxtLogfileName As String
    Public TicksPerLog As Integer
    Public NewLog As Boolean
    Public log_Volts As Double
    Public log_Amps As Double
    Public log_DegC As Double
    Public tickcnt As Integer     ' Count of log ticks

End Structure

Structure CORONALIST
    Public is_valid As Integer
    Public Grp As System.Windows.Forms.GroupBox ' Group to hold corona bar
    Public pb_Corona As System.Windows.Forms.ProgressBar ' pb_Corona_A_0...
End Structure

Public Class Demo_360X

' APP VERSION NUMBER ****************
  Private Const APP_VERSION_MAJOR As Int32 = 2 ' Also update this in AssemblyInfo.vb
  Private Const APP_VERSION_MINOR As Int32 = 2
  Private Const APP_VERSION_BUILD As Int32 = 0
' ***********************************

' NEEDED DLL VERSION NUMBER ****************
  Private Const NEED_DLL_VERSION_MAJOR As Int32 = 4
  Private Const NEED_DLL_VERSION_MINOR As Int32 = 2
  Private Const NEED_DLL_VERSION_BUILD As Int32 = 0
' ***********************************

    Private Const sysid As Int32 = 0
    Private sysinfo As New SYS_INFO
    'Private devcfg As New DEV_CONFIG

    Private syscfg As New SYS_CONFIG
    Private is_valid_syscfg As Int32

    Private Const DEVLIST_MAX = 32
    Private sysdev_list(DEVLIST_MAX) As DEVLIST ' 16 Max = 2-serial * 4-muxserial * 2-dev

    Private Const CORONA_MAX = 16
    Private corona_list(CORONA_MAX) As CORONALIST ' Index same as Accelerator index
    
    Private AMP2SCROLL_BAR_RANGE As Integer = 10000 ' For current setpoints that use amps, convert decimals to/from a fixed-point range for the scrollbars by multiplyng/dividing by this.

    'Private METERTICKS As Integer = 50 ' number of readings to average over
    Private METERTICKS As Integer = 25 ' number of readings to average over
    Private metercnt As Integer = -1
    Private avg_Volts(DEVLIST_MAX, METERTICKS) As Double  ' Volt meter readings to calculate Average
    Private avg_Amps(DEVLIST_MAX, METERTICKS) As Double   ' Current meter readings to calculate Average
    Private avg_DegC(DEVLIST_MAX, METERTICKS) As Double   ' Temperature meter readings to calculate Average

    Private timer_devnum As Int32

    Private ApiVersion As Integer

    Dim init As Integer = 0     ' locks out control handlers

    Private Const SETPOINT_BITS As Integer = 24
    Private Const METER_BITS As Integer = 23

    Private Const FINE_ADJUST_BITS As Integer = 16

    Private SETPOINT_UNITS() As String = {"V", "A"}

    ' Default calibration points expressed as percent/100 of output range
    Private Const CAL_HI_FACTOR As Double = 0.9
    Private Const CAL_LO_FACTOR As Double = 0.4

    Private TICK As Integer = 100 ' timer tick period, ms
    'Private LOGTICKS As Integer = (10000 / TICK) ' number of tick in log period
    'Dim newlog As Integer = 1     ' Start of new log
    'Dim tickcnt As Integer = 0     ' Log time interval

    'Private PASSIVE_BIAS As Integer = 0        ' bias output regulator is active (0) or passive (1)

    'Private EXTRA_PARAMS As Integer = 0        ' show extra parameters in control window (1)
    'Private RESISTOR_CONTROL As Integer = 0    ' bias control uses voltage reg. (0) or variable resistor (1)
    'Private ACC_FILAMENT_FB As Integer = 0     ' use filament to control beam current (1)
    'Private SEPARATED_FEEDBACK As Integer = 0  ' accelerator has a single HV connetion (0) or separated Out and Feedback (1)
    Private CORONA_BY_VOLTAGE As Integer = 0   ' detect corona by current (0) or voltage (1) variations

    '' Test
    'Public TstGrp As System.Windows.Forms.GroupBox
    'Public TstBtn As System.Windows.Forms.Button

    ' =============================== GUI UTILITIES ==============================

    Private Sub ActiveButton(ByVal btn As Button, ByVal active As Boolean)
        Static c As Color
        If active Then
            c = btn.BackColor
            btn.BackColor = Color.Yellow
            btn.Refresh()
        Else
            btn.BackColor = c
        End If
    End Sub

    Private Function LimitText(ByVal t As String) As String
        If (Len(t) < 7) Then Return t
        Return (Val(t) * 1.000001).ToString.Substring(0, 7)
    End Function

    'Private Function IntToHex8(ByVal val As Int32) As String
    '   Return "0x" & RightString("0000000" & Hex$(val), 8)
    'End Function

    Private Function RightString(ByVal s As String, ByVal len As Integer) As String
        Return s.Substring(s.Length - len, len)
    End Function

    ' Find a device in the sysdev_list() that has a windows control with the given name
    Private Function FindDevice(ByRef WinCtrlName As String) As Integer
        ' Look at all controls associated with each device
        For i = 0 To DEVLIST_MAX - 1
            If (sysdev_list(i).valid <> 0) Then
                If (sysdev_list(i).sb_setpoint.Name = WinCtrlName) Or _
                (sysdev_list(i).cb_meteren.Name = WinCtrlName) Or _
                (sysdev_list(i).cb_outen.Name = WinCtrlName) Or _
                (sysdev_list(i).cb_Log.Name = WinCtrlName) _
                Then
                    Return i
                End If
            End If
        Next i
        Return -1
    End Function

    ' ========================================== OPEN/CLOSE ========================================

    Private Sub Demo_360X_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        Dim i As Integer
        ' Test
        'Dim CtlGrp_row As Integer
        'Dim CtlGrp_col As Integer
        'Dim GrpX       As Integer
        'Dim GrpY       As Integer
        'Dim TabStart   As Integer

        Me.ClientSize     = New System.Drawing.Size(432, 480)
        Me.PnlSystem.Size = New System.Drawing.Size(430, 390)

        C3600_ApiOpen(ApiVersion)

        ' List available COM ports
        For Each comport As String In My.Computer.Ports.SerialPortNames
            ComPortList.Items.Add(comport)
        Next
        If ComPortList.Items.Count > 0 Then ComPortList.SelectedIndex = 0

        For i = 0 To DEVLIST_MAX - 1
            sysdev_list(i).valid = 0
        Next

        '-----------------------------------------------------------------
        ' Test 1.
        ' This adds the handler, for the about button, dynamically 
        ' instead of statically at compile time using the 'Handles' directive.
        'AddHandler BtnAbout.Click, AddressOf BtnAbout_Click 
        '-----------------------------------------------------------------

        '-----------------------------------------------------------------
        ' Test 2.
        ' Add group with button to panel. Add handler for button
        'PnlSystem.Enabled = True ' Force for test
        'Me.TstGrp = New System.Windows.Forms.GroupBox
        'Me.TstBtn = New System.Windows.Forms.Button
        'Me.TstGrp.SuspendLayout()
        'Me.PnlSystem.SuspendLayout()
        '' TstGrp
        'Me.TstGrp.Controls.Add(Me.TstBtn)
        'Me.TstGrp.Location = New System.Drawing.Point(428, 3)
        'Me.TstGrp.Name = "GrpTst"
        'Me.TstGrp.Size = New System.Drawing.Size(419, 65)
        'Me.TstGrp.TabIndex = 13
        'Me.TstGrp.TabStop = False
        'Me.TstGrp.Text = "Test"
        '' TstBtn
        'Me.TstBtn.Location = New System.Drawing.Point(4, 14)
        'Me.TstBtn.Name = "BtnTst"
        'Me.TstBtn.Size = New System.Drawing.Size(75, 23)
        'Me.TstBtn.TabIndex = 20
        'Me.TstBtn.Text = "Test"
        'Me.TstBtn.UseVisualStyleBackColor = True
        ''
        'Me.PnlSystem.Controls.Add(Me.TstGrp)
        ''
        'Me.GrpSysConfig.ResumeLayout(False)
        'Me.GrpSysConfig.PerformLayout()
        'Me.ResumeLayout(False)
        ''
        'AddHandler Me.TstBtn.Click, AddressOf BtnAbout_Click 
        'Me.TstGrp.Enabled = True
        '-----------------------------------------------------------------
        'AddCoronaGroup(  0, "_A_0",                  6+0*422, 70+0*92, 15) ' pb_Corona_A_0...
        'AddControlGroup( 0, "_A_0", "Accelerator 1", 6+0*422, 70+1*92, 15, "Voltage to set")
        'AddControlGroup( 1, "_B_0", "Bias 1"       , 6+0*422, 70+2*92, 30, "Voltage to set")
        'AddControlGroup( 2, "_F_0", "Fillament 1"  , 6+0*422, 70+3*92, 45, "Current to set")
        'AddCoronaGroup(  1, "_A_1",                  6+1*422, 70+0*92, 30)
        'AddControlGroup( 3, "_A_1", "Accelerator 2", 6+1*422, 70+1*92, 60, "Voltage to set")
        'AddControlGroup( 4, "_B_1", "Bias 2"       , 6+1*422, 70+2*92, 75, "Voltage to set")
        'AddControlGroup( 5, "_F_1", "Fillament 2"  , 6+1*422, 70+3*92, 90, "Current to set")
                          
        'AddCoronaGroup(  2, "_A_2",                  6+0*422, 70+4*92, 15) ' pb_Corona_A_0...
        'AddControlGroup( 6, "_A_2", "Accelerator 3", 6+0*422, 70+5*92, 15, "Voltage to set")
        'AddControlGroup( 7, "_B_2", "Bias 3"       , 6+0*422, 70+6*92, 30, "Voltage to set")
        'AddControlGroup( 8, "_F_2", "Fillament 3"  , 6+0*422, 70+7*92, 45, "Current to set")
        'AddCoronaGroup(  3, "_A_3",                  6+1*422, 70+4*92, 30)
        'AddControlGroup( 9, "_A_3", "Accelerator 4", 6+1*422, 70+5*92, 60, "Voltage to set")
        'AddControlGroup(10, "_B_3", "Bias 4"       , 6+1*422, 70+6*92, 75, "Voltage to set")
        'AddControlGroup(11, "_F_3", "Fillament 4"  , 6+1*422, 70+7*92, 90, "Current to set")
                          
        'AddCoronaGroup(  4, "_A_4",                  6+0*422, 70+ 8*92, 15) ' pb_Corona_A_0...
        'AddControlGroup(12, "_A_4", "Accelerator 3", 6+0*422, 70+ 9*92, 15, "Voltage to set")
        'AddControlGroup(13, "_B_4", "Bias 3"       , 6+0*422, 70+10*92, 30, "Voltage to set")
        'AddControlGroup(14, "_F_4", "Fillament 3"  , 6+0*422, 70+11*92, 45, "Current to set")
        'CtlGrp_row = 0
        'CtlGrp_col = 0
        'TabStart   = 15
        'For i = 0 To DEVLIST_MAX - 1
        '
        '    GrpX = 6  + CtlGrp_col * 422
        '    GrpY = 70 + CtlGrp_row * 92
        '    ' AddControlGroup(i, "_A_4", "Accelerator 3", GrpX, GrpY, TabStart, "Voltage to set")
        '
        '    TabStart   = TabStart + 15
        '    CtlGrp_row = CtlGrp_row + 1
        '    If ((CtlGrp_row mod 4) = 0) Then
        '        If (CtlGrp_col = 0)
        '            CtlGrp_row = CtlGrp_row - 4 ' Restart rows beside last group of 4
        '            CtlGrp_col = 1 ' = right side
        '        Else 
        '            'CtlGrp_row = start next group below were last group finished
        '            CtlGrp_col = 0 ' = Back to left side
        '        End If
        '    End If
        'Next

        'Grp_A_0.Enabled = False
        'Grp_A_1.Enabled = False
        'Grp_B_0.Enabled = False
        'Grp_B_1.Enabled = False
        'Grp_F_0.Enabled = False
        'Grp_F_1.Enabled = False
        'Grp_Corona_A_0.Enabled = False
        'Grp_Corona_A_1.Enabled = False
        ''pb_Corona_A_0.Value = 0
        ''pb_Corona_A_1.Value = 0

        For i = 0 To CORONA_MAX - 1
            corona_list(i).is_valid = 0
        Next

    End Sub ' Demo_360X_Load

    Private Sub Demo_360X_FormClosing(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles Me.FormClosing
        C3600_ApiClose()
    End Sub

    Private Sub EnableGui(ByVal enable As Boolean)        ' Enable/disable device control
        PnlSystem.Enabled = enable
        BtnResetSystem.Enabled = enable
        ComPortList.Enabled = Not enable
        BtnOpenCloseSys.Text = IIf(enable, "Close", "Open")
        TmrMonitor.Enabled = enable ' And ChkMeterScan_A_0.Checked

        If (enable = False) Then
            'Grp_A_0.Enabled = False
            'Grp_A_1.Enabled = False
            'Grp_B_0.Enabled = False
            'Grp_B_1.Enabled = False
            'Grp_F_0.Enabled = False
            'Grp_F_1.Enabled = False
            'Grp_Corona_A_0.Enabled = False
            'Grp_Corona_A_1.Enabled = False
            ''pb_Corona_A_0.Value = 0
            ''pb_Corona_A_1.Value = 0
            For i = 0 To CORONA_MAX - 1
                If (corona_list(i).is_valid) Then
                    corona_list(i).pb_Corona.Value = 0
                End If
                corona_list(i).is_valid = 0
            Next
        End If

        GprProgress.Enabled = Not enable
        GprProgress.Visible = Not enable
        ProgressBar1.Value = 0 
        ProgressBar2.Value = 0 
        ProgressBar3.Value = 0 

    End Sub

    Private Function OpenSystem() As Boolean
        Dim devcfg As New DEV_CONFIG
        Dim is_valid_devcfg As Int32
        Dim comstr As String
        Dim verstr As String = ""
        Dim comport As Integer
        'Dim sb_max As Integer
        'Dim sb_large_change As Integer

        Dim Grp_Text As String
        Dim DscrSetpoint As String
        Dim GrpX As Integer
        Dim GrpY As Integer
        Dim StrExt As String
        Dim CtlGrp_row As Integer
        Dim CtlGrp_col As Integer
        Dim TabStart As Integer

        ' ApiVersion is a 32-bit API number: major/8, minor/8, build/16.
        If ((NEED_DLL_VERSION_MAJOR <> (ApiVersion >> 24)) or _
            (NEED_DLL_VERSION_MINOR <> ((ApiVersion >> 16) And &HFF)) or _
            (NEED_DLL_VERSION_BUILD <> (ApiVersion And &HFFFF)) ) Then
        
          Dim DLL_Version As String = _
            (ApiVersion >> 24).ToString & "." & _
            ((ApiVersion >> 16) And &HFF).ToString & "." & _
            (ApiVersion And &HFFFF).ToString

          Dim NEED_Version As String = _
            NEED_DLL_VERSION_MAJOR.ToString & "." & _
            NEED_DLL_VERSION_MINOR.ToString & "." & _
            NEED_DLL_VERSION_BUILD.ToString

          MsgBox("Error. Found c3600.dll Ver " & DLL_Version & "  Need Ver " & NEED_Version _
             , MsgBoxStyle.Exclamation, "Error")

          Return False  ' fail
        End If

        comstr = Mid$(ComPortList.SelectedItem, 4)
        comport = Val(comstr) - 1 ' Get host's index for selected serial port
        
        CtlGrp_row = 0
        CtlGrp_col = 0
        TabStart   = 15

        'Dim errcode As Integer = C3600_OpenSystem(sysid, comport, INFINITE_WAIT) ' Open the system
        Dim errcode As Integer = C3600_OpenSystem_with_callback(sysid, comport, INFINITE_WAIT, New OpenSystemCallback(AddressOf OpenSystem_progress) ) ' Open the system with callback
        If (errcode = C3600_ERR_OK) Then
            errcode = C3600_GetSysConfig(sysid, syscfg, is_valid_syscfg, INFINITE_WAIT)  ' Read EEPROM_SYS_REGION data from cluster[0].unit[0] = first device in system (matches ComboDeviceNum.SelectedIndex = 0 below)
            'errcode = C3600_GetDeviceSysConfig(sysid, syscfg, 0, is_valid_syscfg, INFINITE_WAIT)  ' Read EEPROM_SYS_REGION data from cluster[0].unit[0] = first device in system (matches ComboDeviceNum.SelectedIndex = 0 below)

            ' If done here, scrollbars won't automatically show up.
            'If (syscfg.num_clusters > 1) Then
            '    Me.ClientSize     = New System.Drawing.Size(860, 480)
            '    Me.PnlSystem.Size = New System.Drawing.Size(860, 390)
            'Else
            '    Me.ClientSize     = New System.Drawing.Size(432, 480)
            '    Me.PnlSystem.Size = New System.Drawing.Size(430, 390)
            'End If

            If (errcode = C3600_ERR_OK) Then
                SysConfigToGrid()       ' Display system config
                Dim i As Integer = 0    ' Create a combobox entry for each device discovered by the system
                While C3600_GetDevConfig(sysid, i, devcfg, is_valid_devcfg, INFINITE_WAIT) = C3600_ERR_OK

                    ' List all Devices found
                    'cfg->devaddr = (this_devices_clustnum/serialport(0-4) << 3) | (mux_serial_num(0-3) << 1) | dev->unit_devnum (Index of device within its parent unit: 0 = primary device (always implemented), 1 = secondary device (optional).)
                    '
                    ' c.s.d
                    ' c.     clustnum = serialport(0-4) offset
                    '  .s.   Sub-serial port muxed with dtr,rts. (0-3)
                    '  . .d  unit_devnum 0-1 Index of device (EEPROM sector?) within its parent unit
                    '          If unit->has_bridge=1 : 0, 1 Indicates SAD, DEVX for primary (primary's have bridge)
                    '          If unit->has_bridge=0 : 0, 1 Indicates DEV0, DEV1 for slave (no bridge)
                    Dim devaddr As String = _
                       (devcfg.devaddr >> 3).ToString & "." & _
                       ((devcfg.devaddr >> 1) And 3).ToString & "." & _
                       (devcfg.devaddr And 1).ToString

                    verstr = verstr & RightString("0000000" & Hex$(devcfg.fwversion), 4)
                    
                    sysdev_list(i).dev = devcfg

                    StrExt = ""
                    Grp_Text = ""
                    DscrSetpoint = ""
                    If (devcfg.devtype = DEVTYPE.ACCELERATOR) OR _
                       (devcfg.devtype = DEVTYPE.BIAS)        OR _
                       (devcfg.devtype = DEVTYPE.FILAMENT)    Then

                      If (devcfg.devtype = DEVTYPE.ACCELERATOR) Then
                        sysdev_list(i).valid = 1
                        StrExt   = "_A_" & devcfg.type_index
                        Grp_Text = "Accelerator " & devcfg.type_index+1
                        DscrSetpoint = "Voltage to set"
                        GrpX = 6  + CtlGrp_col * 422
                        'GrpY = 70 + CtlGrp_row * 92
                        GrpY = 70 + CtlGrp_row * 92 + 48 ' 48 = 87-39
                        '  corona group = 39 high vs ControlGroup = 87 high
                        AddCoronaGroup(devcfg.type_index, StrExt, GrpX, GrpY, TabStart) ' pb_Corona_A_0...
                        TabStart   = TabStart + 15
                        CtlGrp_row = CtlGrp_row + 1
                        'if (devcfg.devpolarity = 0) Then ' voltage is positive
                        'if (devcfg.devpolarity = 1) Then ' voltage is negative
                      End If
                      If (devcfg.devtype = DEVTYPE.BIAS) Then
                        sysdev_list(i).valid = 1
                        StrExt = "_B_" & devcfg.type_index
                        Grp_Text = "Bias " & devcfg.type_index+1
                        DscrSetpoint = "Voltage to set"
                      End If
                      If (devcfg.devtype = DEVTYPE.FILAMENT) Then
                        sysdev_list(i).valid = 1
                        StrExt = "_F_" & devcfg.type_index
                        Grp_Text = "Fillament " & devcfg.type_index+1
                        DscrSetpoint = "Current to set"
                      End If

                      'AddCoronaGroup(  0, "_A_0",                  6+0*422, 70+0*92, 15) ' pb_Corona_A_0...
                      'AddControlGroup( 0, "_A_0", "Accelerator 1", 6+0*422, 70+1*92, 15, "Voltage to set", is_valid_devcfg)
                      'AddControlGroup( 1, "_B_0", "Bias 1"       , 6+0*422, 70+2*92, 30, "Voltage to set", is_valid_devcfg)
                      'AddControlGroup( 2, "_F_0", "Fillament 1"  , 6+0*422, 70+3*92, 45, "Current to set", is_valid_devcfg)
                      'AddCoronaGroup(  1, "_A_1",                  6+1*422, 70+0*92, 30)
                      'AddControlGroup( 3, "_A_1", "Accelerator 2", 6+1*422, 70+1*92, 60, "Voltage to set", is_valid_devcfg)
                      'AddControlGroup( 4, "_B_1", "Bias 2"       , 6+1*422, 70+2*92, 75, "Voltage to set", is_valid_devcfg)
                      'AddControlGroup( 5, "_F_1", "Fillament 2"  , 6+1*422, 70+3*92, 90, "Current to set", is_valid_devcfg)
                                        
                      'AddCoronaGroup(  2, "_A_2",                  6+0*422, 70+4*92, 15) ' pb_Corona_A_0...
                      'AddControlGroup( 6, "_A_2", "Accelerator 3", 6+0*422, 70+5*92, 15, "Voltage to set", is_valid_devcfg)
                      'AddControlGroup( 7, "_B_2", "Bias 3"       , 6+0*422, 70+6*92, 30, "Voltage to set", is_valid_devcfg)
                      'AddControlGroup( 8, "_F_2", "Fillament 3"  , 6+0*422, 70+7*92, 45, "Current to set", is_valid_devcfg)
                      'AddCoronaGroup(  3, "_A_3",                  6+1*422, 70+4*92, 30)
                      'AddControlGroup( 9, "_A_3", "Accelerator 4", 6+1*422, 70+5*92, 60, "Voltage to set", is_valid_devcfg)
                      'AddControlGroup(10, "_B_3", "Bias 4"       , 6+1*422, 70+6*92, 75, "Voltage to set", is_valid_devcfg)
                      'AddControlGroup(11, "_F_3", "Fillament 4"  , 6+1*422, 70+7*92, 90, "Current to set", is_valid_devcfg)
                                        
                      'AddCoronaGroup(  4, "_A_4",                  6+0*422, 70+ 8*92, 15) ' pb_Corona_A_0...
                      'AddControlGroup(12, "_A_4", "Accelerator 3", 6+0*422, 70+ 9*92, 15, "Voltage to set", is_valid_devcfg)
                      'AddControlGroup(13, "_B_4", "Bias 3"       , 6+0*422, 70+10*92, 30, "Voltage to set", is_valid_devcfg)
                      'AddControlGroup(14, "_F_4", "Fillament 3"  , 6+0*422, 70+11*92, 45, "Current to set", is_valid_devcfg)
                      '  :
                      
                      GrpX = 6  + CtlGrp_col * 422
                      GrpY = 70 + CtlGrp_row * 92
                      AddControlGroup(i, StrExt, Grp_Text, GrpX, GrpY, TabStart, DscrSetpoint, is_valid_devcfg)
                      
                      ' Next Control Group
                      TabStart   = TabStart + 15
                      CtlGrp_row = CtlGrp_row + 1
                      If ((CtlGrp_row mod 4) = 0) Then
                          If (CtlGrp_col = 0)
                              CtlGrp_row = CtlGrp_row - 4 ' Restart rows beside last group of 4
                              CtlGrp_col = 1 ' = right side
                          Else 
                              'CtlGrp_row = start next group below were last group finished
                              CtlGrp_col = 0 ' = Back to left side
                          End If
                      End If
                    End If ' ACCELERATOR or BIAS or FILAMENT

                    '' Save devices configuration, and group it's associated Windows controls with it
                    'If ((devcfg.devtype = DEVTYPE.ACCELERATOR) And devcfg.type_index < 2) Then
                    '    sysdev_list(i).valid = 1
                    '    sysdev_list(i).dev = devcfg
                    '    If (devcfg.type_index = 0) Then
                    '        ChkValidDevConfig_A_0.Checked = is_valid_devcfg
                    '        sysdev_list(i).sb_setpoint = scrollSetpoint_A_0
                    '        If (devcfg.setpoint_type = SIGTYPE.VOLTS) Then
                    '            LblIMeterUnits_A_0.Text = "uA"
                    '            sb_large_change = devcfg.range(devcfg.setpoint_type).max / 100
                    '            sb_max = devcfg.range(devcfg.setpoint_type).max + sb_large_change - 1 ' Max for VOLT device
                    '        Else
                    '            LblIMeterUnits_A_0.Text = "A"
                    '            ' Scrollbar's units are integers, so convert floating point Amp range to larger integer range.
                    '            sb_large_change = devcfg.range(devcfg.setpoint_type).max * AMP2SCROLL_BAR_RANGE / 100
                    '            sb_max = devcfg.range(devcfg.setpoint_type).max * AMP2SCROLL_BAR_RANGE + sb_large_change - 1 ' max for AMP with more integer step range (99 is one less that the LargeChange, to allow SB to go to max value)
                    '        End If
                    '        scrollSetpoint_A_0.Maximum = sb_max
                    '        scrollSetpoint_A_0.LargeChange = sb_large_change
                    '        scrollSetpoint_A_0.SmallChange = 1
                    '        sysdev_list(i).cb_outen = chkOutputEnable_A_0
                    '        sysdev_list(i).cb_meteren = chkMeterScan_A_0
                    '        sysdev_list(i).LblSetpoint = LblSetpoint_A_0
                    '        sysdev_list(i).LblVMeter = LblVMeter_A_0
                    '        sysdev_list(i).LblIMeter = LblIMeter_A_0
                    '        sysdev_list(i).LblTMeter = LblTMeter_A_0
                    '        sysdev_list(i).cb_Log = cb_Log_A_0
                    '        sysdev_list(i).TicksPerLog = 200
                    '        sysdev_list(i).TxtLogfileName = ""
                    '        Grp_A_0.Enabled = True
                    '        'Grp_Corona_A_0.Visible = True ' Default = True
                    '        Grp_Corona_A_0.Enabled = True
                    '        'corona_list(devcfg.type_index) = pb_Corona_A_0
                    '        'corona_list(devcfg.type_index).Value = 0
                    '    ElseIf (devcfg.type_index = 1) Then
                    '        If (syscfg.num_clusters = 1) Then ' Only one SAD per cluster
                    '            sysdev_list(i).valid = 0
                    '        Else
                    '            ChkValidDevConfig_A_1.Checked = is_valid_devcfg
                    '            sysdev_list(i).sb_setpoint = scrollSetpoint_A_1
                    '            If (devcfg.setpoint_type = SIGTYPE.VOLTS) Then
                    '                LblIMeterUnits_A_1.Text = "uA"
                    '                sb_large_change = devcfg.range(devcfg.setpoint_type).max / 100
                    '                sb_max = devcfg.range(devcfg.setpoint_type).max + sb_large_change - 1 ' Max for VOLT device
                    '            Else
                    '                LblIMeterUnits_A_1.Text = "A"
                    '                ' Scrollbar's units are integers, so convert floating point Amp range to larger integer range.
                    '                sb_large_change = devcfg.range(devcfg.setpoint_type).max * AMP2SCROLL_BAR_RANGE / 100
                    '                sb_max = devcfg.range(devcfg.setpoint_type).max * AMP2SCROLL_BAR_RANGE + sb_large_change - 1 ' max for AMP with more integer step range (99 is one less that the LargeChange, to allow SB to go to max value)
                    '            End If
                    '            scrollSetpoint_A_1.Maximum = sb_max
                    '            scrollSetpoint_A_1.LargeChange = sb_large_change
                    '            scrollSetpoint_A_1.SmallChange = 1
                    '            sysdev_list(i).cb_outen = chkOutputEnable_A_1
                    '            sysdev_list(i).cb_meteren = chkMeterScan_A_1
                    '            sysdev_list(i).LblSetpoint = LblSetpoint_A_1
                    '            sysdev_list(i).LblVMeter = LblVMeter_A_1
                    '            sysdev_list(i).LblIMeter = LblIMeter_A_1
                    '            sysdev_list(i).LblTMeter = LblTMeter_A_1
                    '            sysdev_list(i).cb_Log = cb_Log_A_1
                    '            sysdev_list(i).TicksPerLog = 200
                    '            sysdev_list(i).TxtLogfileName = ""
                    '            Grp_A_1.Visible = True
                    '            Grp_A_1.Enabled = True
                    '            Grp_Corona_A_1.Visible = True
                    '            Grp_Corona_A_1.Enabled = True
                    '            'corona_list(devcfg.type_index) = pb_Corona_A_1
                    '            'corona_list(devcfg.type_index).Value = 0
                    '        End If
                    '    End If
                    'ElseIf ((devcfg.devtype = DEVTYPE.BIAS) And devcfg.type_index < 2) Then
                    '    sysdev_list(i).valid = 1
                    '    sysdev_list(i).dev = devcfg
                    '    If (devcfg.type_index = 0) Then
                    '        ChkValidDevConfig_B_0.Checked = is_valid_devcfg
                    '        sysdev_list(i).sb_setpoint = scrollSetpoint_B_0
                    '        If (devcfg.setpoint_type = SIGTYPE.VOLTS) Then
                    '            LblIMeterUnits_B_0.Text = "uA"
                    '            sb_large_change = devcfg.range(devcfg.setpoint_type).max / 100
                    '            sb_max = devcfg.range(devcfg.setpoint_type).max + sb_large_change - 1 ' Max for VOLT device
                    '        Else ' setpoint_type = SIGTYPE.AMPS
                    '            LblIMeterUnits_B_0.Text = "A"
                    '            ' Scrollbar's units are integers, so convert floating point Amp range to larger integer range.
                    '            sb_large_change = devcfg.range(devcfg.setpoint_type).max * AMP2SCROLL_BAR_RANGE / 100
                    '            sb_max = devcfg.range(devcfg.setpoint_type).max * AMP2SCROLL_BAR_RANGE + sb_large_change - 1 ' max for AMP with more integer step range (99 is one less that the LargeChange, to allow SB to go to max value)
                    '        End If
                    '        scrollSetpoint_B_0.Maximum = sb_max
                    '        scrollSetpoint_B_0.LargeChange = sb_large_change
                    '        scrollSetpoint_B_0.SmallChange = 1
                    '        sysdev_list(i).cb_outen = chkOutputEnable_B_0
                    '        sysdev_list(i).cb_meteren = chkMeterScan_B_0
                    '        sysdev_list(i).LblSetpoint = LblSetpoint_B_0
                    '        sysdev_list(i).LblVMeter = LblVMeter_B_0
                    '        sysdev_list(i).LblIMeter = LblIMeter_B_0
                    '        sysdev_list(i).LblTMeter = LblTMeter_B_0
                    '        sysdev_list(i).cb_Log = cb_Log_B_0
                    '        sysdev_list(i).TicksPerLog = 200
                    '        sysdev_list(i).TxtLogfileName = ""
                    '        Grp_B_0.Enabled = True
                    '    ElseIf (devcfg.type_index = 1) Then
                    '        ChkValidDevConfig_B_1.Checked = is_valid_devcfg
                    '        sysdev_list(i).sb_setpoint = scrollSetpoint_B_1
                    '        If (devcfg.setpoint_type = SIGTYPE.VOLTS) Then
                    '            LblIMeterUnits_B_1.Text = "uA"
                    '            sb_large_change = devcfg.range(devcfg.setpoint_type).max / 100
                    '            sb_max = devcfg.range(devcfg.setpoint_type).max + sb_large_change - 1 ' Max for VOLT device
                    '        Else ' setpoint_type = SIGTYPE.AMPS
                    '            LblIMeterUnits_B_1.Text = "A"
                    '            ' Scrollbar's units are integers, so convert floating point Amp range to larger integer range.
                    '            sb_large_change = devcfg.range(devcfg.setpoint_type).max * AMP2SCROLL_BAR_RANGE / 100
                    '            sb_max = devcfg.range(devcfg.setpoint_type).max * AMP2SCROLL_BAR_RANGE + sb_large_change - 1 ' max for AMP with more integer step range (99 is one less that the LargeChange, to allow SB to go to max value)
                    '        End If
                    '        scrollSetpoint_B_1.Maximum = sb_max
                    '        scrollSetpoint_B_1.LargeChange = sb_large_change
                    '        scrollSetpoint_B_1.SmallChange = 1
                    '        sysdev_list(i).cb_outen = chkOutputEnable_B_1
                    '        sysdev_list(i).cb_meteren = chkMeterScan_B_1
                    '        sysdev_list(i).LblSetpoint = LblSetpoint_B_1
                    '        sysdev_list(i).LblVMeter = LblVMeter_B_1
                    '        sysdev_list(i).LblIMeter = LblIMeter_B_1
                    '        sysdev_list(i).LblTMeter = LblTMeter_B_1
                    '        sysdev_list(i).cb_Log = cb_Log_B_1
                    '        sysdev_list(i).TicksPerLog = 200
                    '        sysdev_list(i).TxtLogfileName = ""
                    '        Grp_B_1.Visible = True
                    '        Grp_B_1.Enabled = True
                    '    End If
                    'ElseIf ((devcfg.devtype = DEVTYPE.FILAMENT) And devcfg.type_index < 2) Then
                    '    sysdev_list(i).valid = 1
                    '    sysdev_list(i).dev = devcfg
                    '    If (devcfg.type_index = 0) Then
                    '        ChkValidDevConfig_F_0.Checked = is_valid_devcfg
                    '        sysdev_list(i).sb_setpoint = scrollSetpoint_F_0
                    '        If (devcfg.setpoint_type = SIGTYPE.VOLTS) Then
                    '            LblIMeterUnits_F_0.Text = "uA"
                    '            sb_large_change = devcfg.range(devcfg.setpoint_type).max / 100
                    '            sb_max = devcfg.range(devcfg.setpoint_type).max + sb_large_change - 1 ' Max for VOLT device
                    '        Else ' setpoint_type = SIGTYPE.AMPS
                    '            LblIMeterUnits_F_0.Text = "A"
                    '            ' Scrollbar's units are integers, so convert floating point Amp range to larger integer range.
                    '            sb_large_change = devcfg.range(devcfg.setpoint_type).max * AMP2SCROLL_BAR_RANGE / 100
                    '            sb_max = devcfg.range(devcfg.setpoint_type).max * AMP2SCROLL_BAR_RANGE + sb_large_change - 1 ' max for AMP with more integer step range (99 is one less that the LargeChange, to allow SB to go to max value)
                    '        End If
                    '        scrollSetpoint_F_0.Maximum = sb_max
                    '        scrollSetpoint_F_0.LargeChange = sb_large_change
                    '        scrollSetpoint_F_0.SmallChange = 1
                    '        sysdev_list(i).cb_outen = chkOutputEnable_F_0
                    '        sysdev_list(i).cb_meteren = chkMeterScan_F_0
                    '        sysdev_list(i).LblSetpoint = LblSetpoint_F_0
                    '        sysdev_list(i).LblVMeter = LblVMeter_F_0
                    '        sysdev_list(i).LblIMeter = LblIMeter_F_0
                    '        sysdev_list(i).LblTMeter = LblTMeter_F_0
                    '        sysdev_list(i).cb_Log = cb_Log_F_0
                    '        sysdev_list(i).TicksPerLog = 200
                    '        sysdev_list(i).TxtLogfileName = ""
                    '        Grp_F_0.Enabled = True
                    '    ElseIf (devcfg.type_index = 1) Then
                    '        ChkValidDevConfig_F_1.Checked = is_valid_devcfg
                    '        sysdev_list(i).sb_setpoint = scrollSetpoint_F_1
                    '        If (devcfg.setpoint_type = SIGTYPE.VOLTS) Then
                    '            LblIMeterUnits_F_1.Text = "uA"
                    '            sb_large_change = devcfg.range(devcfg.setpoint_type).max / 100
                    '            sb_max = devcfg.range(devcfg.setpoint_type).max + sb_large_change - 1 ' Max for VOLT device
                    '        Else ' setpoint_type = SIGTYPE.AMPS
                    '            LblIMeterUnits_F_1.Text = "A"
                    '            ' Scrollbar's units are integers, so convert floating point Amp range to larger integer range.
                    '            sb_large_change = devcfg.range(devcfg.setpoint_type).max * AMP2SCROLL_BAR_RANGE / 100
                    '            sb_max = devcfg.range(devcfg.setpoint_type).max * AMP2SCROLL_BAR_RANGE + sb_large_change - 1 ' max for AMP with more integer step range (99 is one less that the LargeChange, to allow SB to go to max value)
                    '        End If
                    '        scrollSetpoint_F_1.Maximum = sb_max
                    '        scrollSetpoint_F_1.LargeChange = sb_large_change
                    '        scrollSetpoint_F_1.SmallChange = 1
                    '        sysdev_list(i).cb_outen = chkOutputEnable_F_1
                    '        sysdev_list(i).cb_meteren = chkMeterScan_F_1
                    '        sysdev_list(i).LblSetpoint = LblSetpoint_F_1
                    '        sysdev_list(i).LblVMeter = LblVMeter_F_1
                    '        sysdev_list(i).LblIMeter = LblIMeter_F_1
                    '        sysdev_list(i).LblTMeter = LblTMeter_F_1
                    '        sysdev_list(i).cb_Log = cb_Log_F_1
                    '        sysdev_list(i).TicksPerLog = 200
                    '        sysdev_list(i).TxtLogfileName = ""
                    '        Grp_F_1.Visible = True
                    '        Grp_F_1.Enabled = True
                    '    End If
                    'End If

                    ' Ver.ACC0:SUB0-SUB1-SUB2-SUB3  ACC1:SUB0-SUB1-SUB2-SUB3
                    If (devcfg.devtype = DEVTYPE.ACCELERATOR) Then
                        verstr = verstr & ":"
                    ElseIf ((devcfg.devtype = DEVTYPE.NONE) And (devcfg.type_index = 1)) Then
                        verstr = verstr & "  "
                    ElseIf Not ((devcfg.devtype = DEVTYPE.NONE) And (devcfg.type_index = 3)) Then
                        verstr = verstr & "-"
                    End If

                    ReadDevSetpoint(i) ' Read setpoint from device and display current state (setpoint + output enable)
                    TmrScrollIdle.Stop() ' Stop pending timer and do it immediately
                    ProgramSetpoint(i)

                    i += 1 ' look for next device

                End While ' While C3600_GetDevConfig()

                'ComboDeviceNum.SelectedIndex = 0 ' Show first device in system

                LblFWVersion.Text = "Ver." & verstr

                ' Re-Size Application window after controls are added.
                ' - If the resizing is done after all the controls are added, the
                '   scrollbars will be displayed when this returns.
                ' - If resizing is done before the controls are added, 
                '  (i.e. before the while loop), no scrollbars are displayed 
                '  until user drags/re-sizes the application window
                If (syscfg.num_clusters > 1) Then
                    Me.ClientSize     = New System.Drawing.Size(860, 480)
                    Me.PnlSystem.Size = New System.Drawing.Size(860, 390)
                Else
                    Me.ClientSize     = New System.Drawing.Size(432, 480)
                    Me.PnlSystem.Size = New System.Drawing.Size(430, 390)
                End If

                'RefreshDev()
                EnableGui(True)      ' Enable device control
                Me.ResumeLayout(True)

                Return True     ' success
            End If
        End If

        'Dim LastError = C3600_HalGetLastError()
        'MsgBox("Error opening system: (" & errcode & ") " & C3600_GetErrorString(errcode) & vbCrLf & _
        '       "LastDllError = " & Err.LastDllError & vbCrLf & _
        '       "LastError = " & LastError & vbCrLf _
        '       , MsgBoxStyle.Exclamation, "Error")
        MsgBox("Error opening system: (" & errcode & ") " & C3600_GetErrorString(errcode) _
               , MsgBoxStyle.Exclamation, "Error")

        Return False  ' fail
    End Function ' OpenSystem

    Private Sub CloseSystem()
        EnableGui(False)            ' disable gui controls
        C3600_CloseSystem(sysid)    ' close the system
        'ComboDeviceNum.Items.Clear()
        For i = 0 To DEVLIST_MAX - 1
            sysdev_list(i).valid = 0
        Next
        'RefreshDev()
        DevEmptyGrid()
        SysEmptyGrid()
    End Sub

    Private Sub OpenCloseSystem(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles BtnOpenCloseSys.Click
        ActiveButton(BtnOpenCloseSys, True)       ' busy button
        If (BtnOpenCloseSys.Text = "Open") Then
            If (Not OpenSystem()) Then
                BtnOpenCloseSys.Text = "Close"
            End If
        Else
            CloseSystem()
        End If

        ActiveButton(BtnOpenCloseSys, False)      ' unbusy button
    End Sub

    Private Sub BtnResetSystem_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles BtnResetSystem.Click
        C3600_SysReset(sysid, 0, INFINITE_WAIT)
        'RefreshDev()
    End Sub

    Private Sub BtnAbout_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles BtnAbout.Click
        'Private Sub BtnAbout_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) ' Test dynamic handler above in Demo_360X_Load()
        ' ApiVersion
        ' Return 32-bit API version number: major/8, minor/8, build/16.
        Dim DLL_Version As String = _
          (ApiVersion >> 24).ToString & "." & _
          ((ApiVersion >> 16) And &HFF).ToString & "." & _
          (ApiVersion And &HFFFF).ToString

        Dim APP_Version As String = _
          APP_VERSION_MAJOR.ToString & "." & _
          APP_VERSION_MINOR.ToString & "." & _
          APP_VERSION_BUILD.ToString

        Dim messageBoxText As String = _
          "CPS" & vbCrLf & _
          "  Demo_360X version : " & APP_Version & vbCrLf & _
          "  mid-3600 DLL API version : " & DLL_Version & vbCrLf
        Dim caption As String = "About"
        Dim button As MessageBoxButtons = MessageBoxButtons.OK
        Dim icon As MessageBoxIcon = MessageBoxIcon.Information
        Dim defaultResult As MessageBoxDefaultButton = MessageBoxDefaultButton.Button1
        'Dim options As Integer = MessageBoxOptions.ServiceNotification
        Dim returnValue As MsgBoxResult

        returnValue = MessageBox.Show(messageBoxText, _
            caption, button, icon, defaultResult, 0) ' options)
    End Sub

    ' ========================================= SYSTEM/DEVICE ATTRIBUTES DISPLAY ====================================

    Private Sub DevEmptyGrid()
        'ChkValidDevConfig_A_0.Checked = False

        ' Device type
        'ComboDevtype.SelectedIndex = 0

    End Sub

    Private Sub SysEmptyGrid()
        ChkValidSysConfig.Checked = False
        txtModel.Text = ""
        txtClusters.Text = ""
        TxtSerialNumber.Text = ""
    End Sub

    Private Sub SysConfigToGrid()
        ChkValidSysConfig.Checked = is_valid_syscfg
        txtModel.Text = syscfg.model
        txtClusters.Text = syscfg.num_clusters
        TxtSerialNumber.Text = syscfg.serial_num
    End Sub

    Private Sub AddControlGroup(ByVal devnum As Integer, StrExt As String, Grp_Text As String, GrpX As Integer, GrpY As Integer, TabNum As Integer, DscrSetpoint As String, is_valid_devcfg As Int32)
      Dim sb_max As Integer
      Dim sb_large_change As Integer
      
      '-------------------------------------------------------------------
      ' Create new controls
      '-------------------------------------------------------------------
      sysdev_list(devnum).Grp = New System.Windows.Forms.GroupBox
      
      sysdev_list(devnum).ChkValidDevConfig = New System.Windows.Forms.CheckBox   ' ChkValidDevConfig_A_0...Checked in OpenSystem() when discovered
      sysdev_list(devnum).cb_outen          = New System.Windows.Forms.CheckBox   ' chkOutputEnable_A_0...
      sysdev_list(devnum).cb_Log            = New System.Windows.Forms.CheckBox   ' cb_Log_A_0...
      sysdev_list(devnum).cb_meteren        = New System.Windows.Forms.CheckBox   ' chkMeterScan_A_0...
      sysdev_list(devnum).sb_setpoint       = New System.Windows.Forms.HScrollBar ' scrollSetpoint_A_0...
      sysdev_list(devnum).LblSetpoint       = New Windows.Forms.Label             ' lblSetpoint_A_0...
      sysdev_list(devnum).LblVMeter         = New Windows.Forms.Label             ' LblVMeter_A_0...
      sysdev_list(devnum).LblTMeter         = New Windows.Forms.Label             ' LblTMeter_A_0...
      sysdev_list(devnum).LblIMeter         = New Windows.Forms.Label             ' LblIMeter_A_0...
      sysdev_list(devnum).DscrSetpoint      = New System.Windows.Forms.Label      ' DscrSetpoint_A_0...   static
      sysdev_list(devnum).DscrVMeter        = New System.Windows.Forms.Label      ' DscrVMeter_A_0...     static
      sysdev_list(devnum).DscrTMeter        = New System.Windows.Forms.Label      ' DscrTMeter_A_0...     static
      sysdev_list(devnum).DscrIMeter        = New System.Windows.Forms.Label      ' DscrIMeter_A_0...     static
      sysdev_list(devnum).LblIMeterUnits    = New System.Windows.Forms.Label      ' LblIMeterUnits_A_0... static / modified in OpenSystem() when discovered

      ' Pause Layout while defining
      sysdev_list(devnum).Grp.SuspendLayout()
      Me.PnlSystem.SuspendLayout()
      Me.SuspendLayout()

      '-------------------------------------------------------------------
      ' Define new controls (Location, Size, etc.)
      ' (From Grp_A_0+ definition's in Demo_360X.Designer.vb)
      '-------------------------------------------------------------------
      sysdev_list(devnum).Grp.Controls.Add(sysdev_list(devnum).ChkValidDevConfig)
      sysdev_list(devnum).Grp.Controls.Add(sysdev_list(devnum).cb_outen)
      sysdev_list(devnum).Grp.Controls.Add(sysdev_list(devnum).cb_Log)
      sysdev_list(devnum).Grp.Controls.Add(sysdev_list(devnum).cb_meteren)
      sysdev_list(devnum).Grp.Controls.Add(sysdev_list(devnum).sb_setpoint)
      sysdev_list(devnum).Grp.Controls.Add(sysdev_list(devnum).LblSetpoint)
      sysdev_list(devnum).Grp.Controls.Add(sysdev_list(devnum).LblVMeter)
      sysdev_list(devnum).Grp.Controls.Add(sysdev_list(devnum).LblTMeter)
      sysdev_list(devnum).Grp.Controls.Add(sysdev_list(devnum).LblIMeter)
      sysdev_list(devnum).Grp.Controls.Add(sysdev_list(devnum).DscrSetpoint)
      sysdev_list(devnum).Grp.Controls.Add(sysdev_list(devnum).DscrVMeter)
      sysdev_list(devnum).Grp.Controls.Add(sysdev_list(devnum).DscrTMeter)
      sysdev_list(devnum).Grp.Controls.Add(sysdev_list(devnum).DscrIMeter)
      sysdev_list(devnum).Grp.Controls.Add(sysdev_list(devnum).LblIMeterUnits)
      ' Grp_A_0...
      sysdev_list(devnum).Grp.Location = New System.Drawing.Point(GrpX, GrpY)
      sysdev_list(devnum).Grp.Name     = "Grp" & StrExt                             ' "Grp_A_0"...
      sysdev_list(devnum).Grp.Size     = New System.Drawing.Size(419, 87)
      sysdev_list(devnum).Grp.TabIndex = TabNum
      sysdev_list(devnum).Grp.TabStop  = False
      sysdev_list(devnum).Grp.Text     = Grp_Text ' "Accelerator 1"...
      ' chkOutputEnable_A_0
      sysdev_list(devnum).cb_outen.AutoSize = True
      sysdev_list(devnum).cb_outen.Location = New System.Drawing.Point(10, 23)
      sysdev_list(devnum).cb_outen.Name = "chkOutputEnable" & StrExt                ' chkOutputEnable_A_0...
      sysdev_list(devnum).cb_outen.Size = New System.Drawing.Size(94, 17)
      sysdev_list(devnum).cb_outen.TabIndex = TabNum+1
      sysdev_list(devnum).cb_outen.Text = "Output Enable"
      sysdev_list(devnum).cb_outen.UseVisualStyleBackColor = True
      ' ChkValidDevConfig_A_0...
      sysdev_list(devnum).ChkValidDevConfig.AutoSize = True
      sysdev_list(devnum).ChkValidDevConfig.Enabled = False
      sysdev_list(devnum).ChkValidDevConfig.Location = New System.Drawing.Point(107, 23)
      sysdev_list(devnum).ChkValidDevConfig.Name = "ChkValidDevConfig" & StrExt     ' ChkValidDevConfig_A_0...
      sysdev_list(devnum).ChkValidDevConfig.Size = New System.Drawing.Size(49, 17)
      sysdev_list(devnum).ChkValidDevConfig.TabIndex = TabNum+2
      sysdev_list(devnum).ChkValidDevConfig.Text = "Valid"
      sysdev_list(devnum).ChkValidDevConfig.UseVisualStyleBackColor = True
      ' DscrSetpoint_A_0
      sysdev_list(devnum).DscrSetpoint.AutoSize = True
      sysdev_list(devnum).DscrSetpoint.Location = New System.Drawing.Point(184, 9)
      sysdev_list(devnum).DscrSetpoint.Name = "DscrSetpoint" & StrExt               ' DscrSetpoint_A_0...
      sysdev_list(devnum).DscrSetpoint.Size = New System.Drawing.Size(72, 13)
      sysdev_list(devnum).DscrSetpoint.TabIndex = TabNum+3
      sysdev_list(devnum).DscrSetpoint.Text = DscrSetpoint
      ' LblSetpoint_A_0
      sysdev_list(devnum).LblSetpoint.AutoSize = True
      sysdev_list(devnum).LblSetpoint.BorderStyle = System.Windows.Forms.BorderStyle.Fixed3D
      sysdev_list(devnum).LblSetpoint.Location = New System.Drawing.Point(184, 25)
      sysdev_list(devnum).LblSetpoint.MinimumSize = New System.Drawing.Size(76, 0)
      sysdev_list(devnum).LblSetpoint.Name = "LblSetpoint" & StrExt                 ' LblSetpoint_A_0
      sysdev_list(devnum).LblSetpoint.Size = New System.Drawing.Size(76, 15)
      sysdev_list(devnum).LblSetpoint.TabIndex = TabNum+4
      sysdev_list(devnum).LblSetpoint.TextAlign = System.Drawing.ContentAlignment.MiddleRight
      ' cb_Log_A_0
      sysdev_list(devnum).cb_Log.AutoSize = True
      sysdev_list(devnum).cb_Log.Location = New System.Drawing.Point(329, 9)
      sysdev_list(devnum).cb_Log.Name = "cb_Log" & StrExt ' ...cb_Log_A_0
      sysdev_list(devnum).cb_Log.Size = New System.Drawing.Size(44, 17)
      sysdev_list(devnum).cb_Log.TabIndex = TabNum+5
      sysdev_list(devnum).cb_Log.Text = "Log"
      sysdev_list(devnum).cb_Log.UseVisualStyleBackColor = True
      ' chkMeterScan_A_0
      sysdev_list(devnum).cb_meteren.AutoSize = True
      sysdev_list(devnum).cb_meteren.Checked = True
      sysdev_list(devnum).cb_meteren.CheckState = System.Windows.Forms.CheckState.Checked
      sysdev_list(devnum).cb_meteren.Location = New System.Drawing.Point(329, 27)
      sysdev_list(devnum).cb_meteren.Name = "chkMeterScan" & StrExt                 ' chkMeterScan_A_0
      sysdev_list(devnum).cb_meteren.Size = New System.Drawing.Size(89, 17)
      sysdev_list(devnum).cb_meteren.TabIndex = TabNum+6
      sysdev_list(devnum).cb_meteren.Text = "Meter Enable"
      sysdev_list(devnum).cb_meteren.UseVisualStyleBackColor = True
      ' scrollSetpoint_A_0
      sysdev_list(devnum).sb_setpoint.LargeChange = 500
      sysdev_list(devnum).sb_setpoint.Location = New System.Drawing.Point(5, 45)
      sysdev_list(devnum).sb_setpoint.Maximum = 5000
      sysdev_list(devnum).sb_setpoint.Name = "scrollSetpoint" & StrExt              ' scrollSetpoint_A_0
      sysdev_list(devnum).sb_setpoint.Size = New System.Drawing.Size(409, 19)
      sysdev_list(devnum).sb_setpoint.TabIndex = TabNum+7
      ' DscrVMeter_A_0
      sysdev_list(devnum).DscrVMeter.AutoSize = True
      sysdev_list(devnum).DscrVMeter.Location = New System.Drawing.Point(7, 69)
      sysdev_list(devnum).DscrVMeter.Name = "DscrVMeter" & StrExt                   ' DscrVMeter_A_0
      sysdev_list(devnum).DscrVMeter.Size = New System.Drawing.Size(72, 13)
      sysdev_list(devnum).DscrVMeter.TabIndex = TabNum+8
      sysdev_list(devnum).DscrVMeter.Text = "Voltage meter"
      ' LblVMeter_A_0
      sysdev_list(devnum).LblVMeter.AutoSize = True
      sysdev_list(devnum).LblVMeter.BorderStyle = System.Windows.Forms.BorderStyle.Fixed3D
      sysdev_list(devnum).LblVMeter.Location = New System.Drawing.Point(79, 69)
      sysdev_list(devnum).LblVMeter.MinimumSize = New System.Drawing.Size(76, 0)
      sysdev_list(devnum).LblVMeter.Name = "LblVMeter" & StrExt                     ' LblVMeter_A_0
      sysdev_list(devnum).LblVMeter.Size = New System.Drawing.Size(76, 15)
      sysdev_list(devnum).LblVMeter.TabIndex = TabNum+9
      ' DscrTMeter_A_0
      sysdev_list(devnum).DscrTMeter.AutoSize = True
      sysdev_list(devnum).DscrTMeter.Location = New System.Drawing.Point(169, 69)
      sysdev_list(devnum).DscrTMeter.Name = "DscrTMeter" & StrExt                   ' DscrTMeter_A_0
      sysdev_list(devnum).DscrTMeter.Size = New System.Drawing.Size(18, 13)
      sysdev_list(devnum).DscrTMeter.TabIndex = TabNum+10
      sysdev_list(devnum).DscrTMeter.Text = "°C"
      ' LblTMeter_A_0
      sysdev_list(devnum).LblTMeter.AutoSize = True
      sysdev_list(devnum).LblTMeter.BorderStyle = System.Windows.Forms.BorderStyle.Fixed3D
      sysdev_list(devnum).LblTMeter.Location = New System.Drawing.Point(189, 69)
      sysdev_list(devnum).LblTMeter.MinimumSize = New System.Drawing.Size(76, 0)
      sysdev_list(devnum).LblTMeter.Name = "LblTMeter" & StrExt                     ' LblTMeter_A_0
      sysdev_list(devnum).LblTMeter.Size = New System.Drawing.Size(76, 15)
      sysdev_list(devnum).LblTMeter.TabIndex = TabNum+11
      ' DscrIMeter_A_0
      sysdev_list(devnum).DscrIMeter.AutoSize = True
      sysdev_list(devnum).DscrIMeter.Location = New System.Drawing.Point(274, 69)
      sysdev_list(devnum).DscrIMeter.Name = "DscrIMeter" & StrExt                   ' DscrIMeter_A_0
      sysdev_list(devnum).DscrIMeter.Size = New System.Drawing.Size(41, 13)
      sysdev_list(devnum).DscrIMeter.TabIndex = TabNum+12
      sysdev_list(devnum).DscrIMeter.Text = "Current"
      ' LblIMeter_A_0
      sysdev_list(devnum).LblIMeter.AutoSize = True
      sysdev_list(devnum).LblIMeter.BorderStyle = System.Windows.Forms.BorderStyle.Fixed3D
      sysdev_list(devnum).LblIMeter.Location = New System.Drawing.Point(319, 69)
      sysdev_list(devnum).LblIMeter.MinimumSize = New System.Drawing.Size(76, 0)
      sysdev_list(devnum).LblIMeter.Name = "LblIMeter" & StrExt                     ' LblIMeter_A_0
      sysdev_list(devnum).LblIMeter.Size = New System.Drawing.Size(76, 15)
      sysdev_list(devnum).LblIMeter.TabIndex = TabNum+13
      ' LblIMeterUnits_A_0
      sysdev_list(devnum).LblIMeterUnits.AutoSize = True
      sysdev_list(devnum).LblIMeterUnits.Location = New System.Drawing.Point(398, 69)
      sysdev_list(devnum).LblIMeterUnits.Name = "LblIMeterUnits" & StrExt           ' LblIMeterUnits_A_0
      sysdev_list(devnum).LblIMeterUnits.Size = New System.Drawing.Size(14, 13)
      sysdev_list(devnum).LblIMeterUnits.TabIndex = TabNum+14
      sysdev_list(devnum).LblIMeterUnits.Text = "A"

      ' Add group to pannel
      Me.PnlSystem.Controls.Add(sysdev_list(devnum).Grp)

      ' Resume Layout
      sysdev_list(devnum).Grp.ResumeLayout(False)
      sysdev_list(devnum).Grp.PerformLayout()

      Me.PnlSystem.ResumeLayout(False)
      Me.ResumeLayout(False)
      
      AddHandler sysdev_list(devnum).cb_Log.CheckedChanged,    AddressOf cb_Log_CheckedChanged
      AddHandler sysdev_list(devnum).sb_setpoint.ValueChanged, AddressOf SetpointScrollChanged
      AddHandler sysdev_list(devnum).cb_outen.CheckedChanged,  AddressOf chkOutputEnableChanged

      ' This stuff used to be done in OpenSystem()
      'If (sysdev_list(devnum).dev.devtype = DEVTYPE.ACCELERATOR) OR _
      '   (sysdev_list(devnum).dev.devtype = DEVTYPE.BIAS) Then
        sysdev_list(devnum).ChkValidDevConfig.Checked = is_valid_devcfg ' ChkValidDevConfig_A_0.Checked = is_valid_devcfg
        If (sysdev_list(devnum).dev.setpoint_type = SIGTYPE.VOLTS) Then
            sysdev_list(devnum).LblIMeterUnits.Text = "uA" ' LblIMeterUnits_A_0.Text = "uA"
            sb_large_change = sysdev_list(devnum).dev.range(sysdev_list(devnum).dev.setpoint_type).max / 100
            sb_max = sysdev_list(devnum).dev.range(sysdev_list(devnum).dev.setpoint_type).max + sb_large_change - 1 ' Max for VOLT device
        Else
            sysdev_list(devnum).LblIMeterUnits.Text = "A"  ' LblIMeterUnits_A_0.Text = "A"
            ' Scrollbar's units are integers, so convert floating point Amp range to larger integer range.
            sb_large_change = (sysdev_list(devnum).dev.range(sysdev_list(devnum).dev.setpoint_type).max * AMP2SCROLL_BAR_RANGE) / 100
            sb_max = (sysdev_list(devnum).dev.range(sysdev_list(devnum).dev.setpoint_type).max * AMP2SCROLL_BAR_RANGE) + sb_large_change - 1 ' max for AMP with more integer step range (99 is one less that the LargeChange, to allow SB to go to max value)
        End If
        sysdev_list(devnum).sb_setpoint.Maximum = sb_max               ' scrollSetpoint_A_0.Maximum = sb_max
        sysdev_list(devnum).sb_setpoint.LargeChange = sb_large_change  ' scrollSetpoint_A_0.LargeChange = sb_large_change
        sysdev_list(devnum).sb_setpoint.SmallChange = 1                ' scrollSetpoint_A_0.SmallChange = 1
        sysdev_list(devnum).TicksPerLog = 200
        sysdev_list(devnum).TxtLogfileName = ""
        sysdev_list(devnum).Grp.Enabled = True ' Grp_A_0.Enabled = True
      'End If
      
    End Sub ' AddControlGroup
   'AddControlGroup(0, "_A_0", "Accelerator 1", 6+0*422, 70+1*92, 15, "Voltage to set")
   'AddControlGroup(1, "_B_0", "Bias 1"       , 6+0*422, 70+2*92, 30, "Voltage to set")
   'AddControlGroup(2, "_F_0", "Fillament 1"  , 6+0*422, 70+3*92, 45, "Current to set")
   'AddControlGroup(3, "_A_1", "Accelerator 2", 6+1*422, 70+1*92, 60, "Voltage to set")
   'AddControlGroup(4, "_B_1", "Bias 2"       , 6+1*422, 70+2*92, 75, "Voltage to set")
   'AddControlGroup(5, "_F_1", "Fillament 2"  , 6+1*422, 70+3*92, 90, "Current to set")

    Private Sub AddCoronaGroup(ByVal coronanum As Integer, StrExt As String, GrpX As Integer, GrpY As Integer, TabNum As Integer)
      '-------------------------------------------------------------------
      ' Create new controls
      '-------------------------------------------------------------------
      'Me.Grp_Corona_A_0 = New System.Windows.Forms.GroupBox
      'Me.Grp_Corona_A_1 = New System.Windows.Forms.GroupBox
      'Me.pb_Corona_A_1 = New System.Windows.Forms.ProgressBar
      'Me.pb_Corona_A_0 = New System.Windows.Forms.ProgressBar
      corona_list(coronanum).Grp = New System.Windows.Forms.GroupBox          ' Grp_Corona_A_0...
      corona_list(coronanum).pb_Corona = New System.Windows.Forms.ProgressBar ' pb_Corona_A_0...
      corona_list(coronanum).is_valid = 1

      ' Pause Layout while defining
      corona_list(coronanum).Grp.SuspendLayout()
      Me.PnlSystem.SuspendLayout()
      Me.SuspendLayout()
      
      '-------------------------------------------------------------------
      ' Define new controls
      ' (From Grp_A_0+ definition's in Demo_360X.Designer.vb)
      '-------------------------------------------------------------------
      'Grp_Corona_A_0                                                              'Grp_Corona_A_1
      corona_list(coronanum).Grp.Controls.Add(corona_list(coronanum).pb_Corona)    'Me.Grp_Corona_A_1.Controls.Add(Me.pb_Corona_A_1)
      corona_list(coronanum).Grp.Location = New System.Drawing.Point(GrpX, GrpY)   'Me.Grp_Corona_A_1.Location = New System.Drawing.Point(428, 343)
      corona_list(coronanum).Grp.Name = "Grp_Corona" & StrExt 'Grp_Corona_A_0      'Me.Grp_Corona_A_1.Name = "Grp_Corona_A_1"
      corona_list(coronanum).Grp.Size = New System.Drawing.Size(419, 39)           'Me.Grp_Corona_A_1.Size = New System.Drawing.Size(419, 39)
      corona_list(coronanum).Grp.TabIndex = TabNum+0                               'Me.Grp_Corona_A_1.TabIndex = 46
      corona_list(coronanum).Grp.TabStop = False                                   'Me.Grp_Corona_A_1.TabStop = False
      corona_list(coronanum).Grp.Text = "Corona detector"                          'Me.Grp_Corona_A_1.Text = "Corona detector"
      corona_list(coronanum).Grp.Visible = True                                    'Me.Grp_Corona_A_1.Visible = False
      'corona_list(coronanum).Grp.Enabled = True                                   ' Grp_Corona_A_0.Enabled = True

      
      'pb_Corona_A_0                                                               'pb_Corona_A_1
      corona_list(coronanum).pb_Corona.Location = New System.Drawing.Point(14, 18) 'Me.pb_Corona_A_1.Location = New System.Drawing.Point(14, 18)
      corona_list(coronanum).pb_Corona.Maximum = 1000                              'Me.pb_Corona_A_1.Maximum = 1000
      corona_list(coronanum).pb_Corona.Name = "pb_Corona" & StrExt ' pb_Corona_A_0 'Me.pb_Corona_A_1.Name = "pb_Corona_A_1"
      corona_list(coronanum).pb_Corona.Size = New System.Drawing.Size(390, 10)     'Me.pb_Corona_A_1.Size = New System.Drawing.Size(390, 10)
      corona_list(coronanum).pb_Corona.TabIndex = TabNum+1                         'Me.pb_Corona_A_1.TabIndex = 44

      ' Add group to pannel
      Me.PnlSystem.Controls.Add(corona_list(coronanum).Grp)

      ' Resume Layout
      corona_list(coronanum).Grp.ResumeLayout(False)
      corona_list(coronanum).Grp.PerformLayout()

      Me.PnlSystem.ResumeLayout(False)
      Me.ResumeLayout(False)

      corona_list(coronanum).Grp.Visible = True ' Grp_Corona_A_0.Visible = True
      corona_list(coronanum).Grp.Enabled = True ' Grp_Corona_A_0.Enabled = True
    End Sub ' AddCoronaGroup
   'AddCoronaGroup(0, "_A_0", 6+0*422, 70+0*92, 15)
   'AddCoronaGroup(1, "_A_1", 6+1*422, 70+0*92, 30)

    ' ============================================= SYSTEM/DEVICE CONFIGURATION ================================================

    ' Select device and display its configuration
    'Private Sub DeviceNumChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ComboDeviceNum.SelectedIndexChanged
    '    ' Do nothing. Just show user all the devices
    'End Sub

    'Private Sub cb_Log_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cb_Log_A_0.CheckedChanged, cb_Log_B_0.CheckedChanged, cb_Log_F_0.CheckedChanged, cb_Log_A_1.CheckedChanged, cb_Log_B_1.CheckedChanged, cb_Log_F_1.CheckedChanged
    Private Sub cb_Log_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs)
        Dim devnum As Integer
        Dim setupLog As New setup_log
        Dim returnValue As MsgBoxResult

        devnum = FindDevice(sender.name())

        ' If (cb_Log_A_0.Checked = True) Then
        If (sysdev_list(devnum).cb_Log.Checked = True) Then

            setupLog.ChkEnableLog.Checked = sysdev_list(devnum).EnableLog
            setupLog.TxtLogfileName.Text = sysdev_list(devnum).TxtLogfileName
            setupLog.TicksPerLog.Text = sysdev_list(devnum).TicksPerLog
            returnValue = setupLog.ShowDialog()
            If returnValue = DialogResult.OK Then
                'MessageBox.Show("OK" & vbCrLf & _
                '                "EnableLog = " & setupLog.ChkEnableLog.Checked.ToString & vbCrLf & _
                '                "Log file name = " & setupLog.TxtLogfileName.Text & vbCrLf & _
                '                "Ticks per log = " & setupLog.TicksPerLog.Text & vbCrLf _
                '                , "Caption", MessageBoxButtons.OK)

                ' Save DialogBox settings regaurdless
                sysdev_list(devnum).EnableLog = setupLog.ChkEnableLog.Checked
                sysdev_list(devnum).TxtLogfileName = setupLog.TxtLogfileName.Text
                sysdev_list(devnum).TicksPerLog = setupLog.TicksPerLog.Text

                If (setupLog.ChkEnableLog.Checked = True) Then
                    sysdev_list(devnum).NewLog = True
                    sysdev_list(devnum).tickcnt = 0
                Else
                    ' Cancel GUI checkbox since user didn't enable it in the dialog box
                    sysdev_list(devnum).cb_Log.Checked = False
                End If
            Else
                ' Cancel GUI checkbox since user canceled the dialog box
                'MessageBox.Show("Cancel", "Caption", MessageBoxButtons.OK)
                sysdev_list(devnum).cb_Log.Checked = False
            End If
        Else
            sysdev_list(devnum).EnableLog = False
        End If

    End Sub

    ' ================================== Logging =====================================================

    Private Sub LogLine(ByRef txt As String, ByRef FileName As String)
        ' Print txt to log file
        'My.Computer.FileSystem.WriteAllText(TxtLogfileName.Text, txt & vbCrLf, True)
        Dim sw As StreamWriter = File.AppendText(FileName)
        sw.WriteLine(txt)
        sw.Flush()
        sw.Close()
    End Sub

    Private Sub LogToFile(ByVal devnum As Integer, ByVal Volts As Double, ByVal Amps As Double, ByVal DegC As Integer)
        'Static Dim log_Volts As Double
        'Static Dim log_Amps As Double
        'Static Dim log_DegC As Double

        Dim avg_Volts As Double
        Dim avg_Amps As Double
        Dim avg_DegC As Double

        If (sysdev_list(devnum).NewLog = True) Then
            Dim time As Date = TimeOfDay
            Dim hdr As String
            Dim SecPerLog As Integer

            If (sysdev_list(devnum).TxtLogfileName.Length = 0) Then
                sysdev_list(devnum).cb_Log.Checked = False
                Return
            End If

            If File.Exists(sysdev_list(devnum).TxtLogfileName) = False Then
                ' No file? Create it
                Dim sw As StreamWriter = File.CreateText(sysdev_list(devnum).TxtLogfileName)
                hdr = time
                hdr.PadRight(25, " "c)
                'sysdev_list(devnum).TicksPerLog                  = 100 ticks per log
                'sysdev_list(devnum).TicksPerLog * TICK / 1000    = 100t/l * 100ms/t / 1000ms/s = 100*100/1000 s/l
                TICK = TmrMonitor.Interval ' in milliseconds
                'LOGTICKS = sysdev_list(devnum).TicksPerLog
                SecPerLog = sysdev_list(devnum).TicksPerLog * TICK / 1000
                hdr &= sysdev_list(devnum).TicksPerLog & " times averaged, " & SecPerLog & "seconds period"
                'sw.WriteLine(hdr)
                sw.Flush()
                sw.Close()
            End If
            sysdev_list(devnum).NewLog = False
            sysdev_list(devnum).log_Volts = 0
            sysdev_list(devnum).log_Amps = 0
            sysdev_list(devnum).log_DegC = 0
        End If

        ' Accumulate samples every tick
        sysdev_list(devnum).log_Volts = sysdev_list(devnum).log_Volts + Volts
        sysdev_list(devnum).log_Amps = sysdev_list(devnum).log_Amps + Amps
        sysdev_list(devnum).log_DegC = sysdev_list(devnum).log_DegC + DegC

        If (sysdev_list(devnum).tickcnt < sysdev_list(devnum).TicksPerLog - 1) Then
            sysdev_list(devnum).tickcnt = sysdev_list(devnum).tickcnt + 1
        Else ' tickcnt >= sysdev_list(devnum).TicksPerLog
            Dim txt As String
            'txt = String.Format("{0,10:#.##} {0,12:#.###}", Volts, Amps * 1000000)

            avg_Volts = sysdev_list(devnum).log_Volts / sysdev_list(devnum).TicksPerLog
            avg_Amps = sysdev_list(devnum).log_Amps / sysdev_list(devnum).TicksPerLog
            avg_DegC = sysdev_list(devnum).log_DegC / sysdev_list(devnum).TicksPerLog

            txt = String.Format("{0,10:#.##} {1,12:#.#########}", avg_Volts, avg_Amps)
            LogLine(txt, sysdev_list(devnum).TxtLogfileName)
            sysdev_list(devnum).tickcnt = 0
            sysdev_list(devnum).log_Volts = 0
            sysdev_list(devnum).log_Amps = 0
            sysdev_list(devnum).log_DegC = 0
        End If
    End Sub

    ' ================================== METERING =====================================================

    '-----------------------------------------------------------------
    ' Calculate averages for meter readings
    '-----------------------------------------------------------------
    Private Sub CalcAverageMeterReadings(ByVal devnum As Integer, ByRef sysdev As DEVLIST, ByVal Volts As Double, ByVal Amps As Double, ByVal DegC As Double, ByRef out_Volts As Double, ByRef out_Amps As Double, ByRef out_DegC As Double)
        'Dim d As Integer
        'Dim i As Integer

        '-----------------------------------------------------------------
        ' Display immediate readings
        '-----------------------------------------------------------------
        'out_Volts = Volts
        'out_Amps = Amps
        'out_DegC = DegC

        '-----------------------------------------------------------------
        ' For each device
        '-----------------------------------------------------------------
        If (metercnt = -1) Then
            'For d = 0 To DEVLIST_MAX - 1 ' all devices (whether their valid or not)
            For m = 0 To METERTICKS - 1
                ' This takes a wile to build up to the average value...
                'avg_Volts(d, m) = 0
                'avg_Amps(d, m) = 0
                'avg_DegC(d, m) = 0
                'sysdev.avgVMeter = 0
                'sysdev.avgIMeter = 0
                'sysdev.avgTMeter = 0

                ' Set initial average to first reading
                avg_Volts(devnum, m) = Volts
                avg_Amps(devnum, m) = Amps
                avg_DegC(devnum, m) = DegC
            Next m
            'Next d
            ' Set initial average to first reading
            sysdev.avgVMeter = Volts * METERTICKS
            sysdev.avgIMeter = Amps * METERTICKS
            sysdev.avgTMeter = DegC * METERTICKS
            'metercnt = 0 ' done in caller
        Else ' metercnt = 0 - METERTICKS-1
            ' Make average calculation faster
            ' avgtotal = avgtotal + new - old
            sysdev.avgVMeter = sysdev.avgVMeter + Volts - avg_Volts(devnum, metercnt)
            sysdev.avgIMeter = sysdev.avgIMeter + Amps - avg_Amps(devnum, metercnt)
            sysdev.avgTMeter = sysdev.avgTMeter + DegC - avg_DegC(devnum, metercnt)

            ' Update new value
            avg_Volts(devnum, metercnt) = Volts
            avg_Amps(devnum, metercnt) = Amps
            avg_DegC(devnum, metercnt) = DegC
            'Volts = Amps = DegC = 0
            'For i = 0 To METERTICKS - 1
            '    Volts = Volts + avg_Volts(devnum, metercnt)
            '    Amps = Amps + avg_Amps(devnum, metercnt)
            '    DegC = DegC + avg_DegC(devnum, metercnt)
            'Next i
            'out_Volts = Volts / METERTICKS
            'out_Amps = Amps / METERTICKS
            'out_DegC = DegC / METERTICKS
        End If

        out_Volts = sysdev.avgVMeter / METERTICKS
        out_Amps = sysdev.avgIMeter / METERTICKS
        out_DegC = sysdev.avgTMeter / METERTICKS
    End Sub

    Private Function detcorona(ByRef sysdev As DEVLIST, ByVal i_or_v As Double) As Double
        Static tconst As Double = 5
        'Static lasti As Double
        'Static lastdi1 As Double
        'Static det As Double
        Dim nlim, plim, di1, di2, dd As Double

        If (i_or_v = 0) Then
            sysdev.det = 0
            Return sysdev.det
        End If
        If (i_or_v = sysdev.lasti) Then
            Return sysdev.det
        End If

        plim = i_or_v * 0.1 ' +10%
        nlim = -plim        ' -10%

        di1 = i_or_v - sysdev.lasti ' delta
        If (((di1 > 0) And (sysdev.lastdi1 > 0)) Or ((di1 < 0) And (sysdev.lastdi1 < 0))) Then
            'di1 = 0;
            di2 = 0
        Else ' +di1 & -lastdi1  or -di1 & +lastdi1
            If (di1 > plim) Then
                di1 = plim
            ElseIf (di1 < nlim) Then
                di1 = nlim
            End If

            di2 = di1 - sysdev.lastdi1
            If (di2 > plim) Then
                di2 = plim
            ElseIf (di2 < nlim) Then
                di2 = nlim
            End If
        End If

        dd = Math.Abs(di2) - sysdev.det
        sysdev.det += dd / tconst

        sysdev.lasti = i_or_v
        sysdev.lastdi1 = di1

        Return sysdev.det
    End Function

    ' On each TmrMonitor.Tick Read each enabled devices meters and update it's window controls
    Private Sub TmrMonitor_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles TmrMonitor.Tick
        Dim devcfg As New DEV_CONFIG
        Dim Volts As Double
        Dim Amps As Double
        Dim DegC As Double
        Dim mtr_Volts As Double
        Dim mtr_Amps As Double
        Dim mtr_DegC As Double
        Dim RawVolts As Integer
        Dim RawAmps As Integer
        Dim dCoronalevel As Double
        Dim iCoronalevel As Integer

        Dim i As Integer

        For i = 0 To DEVLIST_MAX - 1
            If (sysdev_list(i).valid <> 0) Then
                ' Display averaged meter readings
                If (sysdev_list(i).cb_meteren.Checked) Then
                    devcfg = sysdev_list(i).dev
                    C3600_GetMeter(sysid, devcfg.devtype, devcfg.type_index, SIGTYPE.VOLTS, Volts, RawVolts, INFINITE_WAIT)
                    C3600_GetMeter(sysid, devcfg.devtype, devcfg.type_index, SIGTYPE.AMPS, Amps, RawAmps, INFINITE_WAIT)
                    C3600_GetMeter(sysid, devcfg.devtype, devcfg.type_index, METER_T, DegC, IntPtr.Zero, INFINITE_WAIT)

                    CalcAverageMeterReadings(i, sysdev_list(i), Volts, Amps, DegC, mtr_Volts, mtr_Amps, mtr_DegC)

                    If (sysdev_list(i).dev.setpoint_type = SIGTYPE.VOLTS) Then
                        mtr_Amps = mtr_Amps * 1000000 ' Display uA for volt meters on voltage supplies
                    End If
                    sysdev_list(i).LblVMeter.Text = LimitText(mtr_Volts.ToString) ' LblVMeter_A_0...
                    sysdev_list(i).LblIMeter.Text = LimitText(mtr_Amps.ToString) ' LblIMeter_A_0...
                    sysdev_list(i).LblTMeter.Text = LimitText(mtr_DegC.ToString) ' LblTMeter_A_0

                    If (sysdev_list(i).EnableLog = True) Then
                        LogToFile(i, mtr_Volts, mtr_Amps, mtr_DegC)
                    Else
                        sysdev_list(i).NewLog = True
                        sysdev_list(i).tickcnt = 0
                    End If
                End If ' meter

                ' Calculate and display corona level for Accelerator's
                If (devcfg.devtype = DEVTYPE.ACCELERATOR) And (sysdev_list(i).cb_outen.Checked = True) Then
                    ' enabled acceerator
                    If (CORONA_BY_VOLTAGE) Then
                        dCoronalevel = detcorona(sysdev_list(i), Volts) * 100         ' full scale (1000 units) is 10V
                    Else
                        dCoronalevel = detcorona(sysdev_list(i), Amps) * 1000000000.0 ' full scale (1000 units) is 1 uA
                    End If
                    iCoronalevel = dCoronalevel
                    iCoronalevel = Math.Max(0, iCoronalevel)
                    iCoronalevel = Math.Min(1000, iCoronalevel)
                    'If (devcfg.type_index = 0) Then
                    '    pb_Corona_A_0.Value = iCoronalevel
                    'Else
                    '    pb_Corona_A_1.Value = iCoronalevel
                    'End If
                    corona_list(devcfg.type_index).pb_Corona.Value = iCoronalevel
                    
                End If ' Corona
            End If ' device valid

        Next ' all devices

        If (metercnt = -1) Then ' Done below via inc
            metercnt = 0
        End If

        metercnt = metercnt + 1 ' Used in CalcAverageMeterReadings() but updated here after all readings done
        If (metercnt >= METERTICKS) Then
            metercnt = 0
        End If

    End Sub

    '================================= SETPOINT & OUTPUT ENABLE ======================================

    ' Set in:
    '   ReadDevSetpoint
    Private Property SetpointScrollValue(ByVal devnum As Integer) As Integer
        Get
            'Return scrollSetpoint_A_0.Value
            Return sysdev_list(devnum).sb_setpoint.Value
        End Get
        Set(ByVal setpoint As Integer)
            init += 1
            If setpoint < 0 Then
                setpoint = 0
            End If
            'scrollSetpoint_A_0.Value = setpoint ' set slider
            sysdev_list(devnum).sb_setpoint.Value = setpoint
            init -= 1
            'LblSetpoint_A_0.Text = setpoint.ToString
            If (sysdev_list(devnum).dev.setpoint_type = SIGTYPE.AMPS) Then
                setpoint = setpoint / AMP2SCROLL_BAR_RANGE ' ((0-max)*10000)/10000 = Amps....scrolbar range = 0 to max*10000
            End If
            sysdev_list(devnum).LblSetpoint.Text = setpoint.ToString
        End Set
    End Property

    'Private Sub SetpointScrollChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles scrollSetpoint_A_0.ValueChanged, scrollSetpoint_B_0.ValueChanged, scrollSetpoint_F_0.ValueChanged, scrollSetpoint_A_1.ValueChanged, scrollSetpoint_B_1.ValueChanged, scrollSetpoint_F_1.ValueChanged
    Private Sub SetpointScrollChanged(ByVal sender As System.Object, ByVal e As System.EventArgs)
        Dim devnum As Integer
        Dim NewSetpoint As Double

        devnum = FindDevice(sender.name())
        If ((devnum >= 0) and (devnum < DEVLIST_MAX)) Then
          timer_devnum = devnum
          
          ' generic for all scrollbars and their text label
          'sysdev_list(devnum).LblSetpoint.Text = sysdev_list(devnum).sb_setpoint.Value.ToString
          NewSetpoint = sysdev_list(devnum).sb_setpoint.Value
          If (sysdev_list(devnum).dev.setpoint_type = SIGTYPE.AMPS) Then
              NewSetpoint = NewSetpoint / AMP2SCROLL_BAR_RANGE ' ((0-max)*10000)/10000 = Amps....scrolbar range = 0 to max*10000
          End If
          sysdev_list(devnum).LblSetpoint.Text = NewSetpoint.ToString
          
          ' Start timer if it's not already running, and program the setpoint after a delay in ProgramSetpoint()
          If (init = 0) Then 
            TmrScrollIdle.Start()
          End If
        Else
          MsgBox("Error Finding control: (" & sender.name() & ") "  _
               , MsgBoxStyle.Exclamation, "Error")
        End If
    End Sub

    ' Read current setpoint and output enable for given device
    ' - Indirectly programs setpoint by changing the scrollbar
    Private Function ReadDevSetpoint(ByVal devnum As Integer) As Int32
        Dim raw As Integer
        Dim setpoint As Double
        Dim enable As Integer
        Dim errcode As Int32
        Dim devtype As DEVTYPE
        Dim type_index As Integer

        setpoint = 0
        If (sysdev_list(devnum).valid <> 0) Then
            devtype = sysdev_list(devnum).dev.devtype
            type_index = sysdev_list(devnum).dev.type_index

            errcode = C3600_GetOutput(sysid, devtype, type_index, setpoint, raw, enable, INFINITE_WAIT)

            If (sysdev_list(devnum).dev.setpoint_type = SIGTYPE.AMPS) Then
                setpoint = setpoint * AMP2SCROLL_BAR_RANGE
            End If

            SetpointScrollValue(devnum) = setpoint ' set devices scrollbar and text lable
            init += 1

            'chkOutputEnable_A_0.Checked = (enable <> 0)
            sysdev_list(devnum).cb_outen.Checked = (enable <> 0)
            init -= 1
        End If
        Return setpoint
    End Function


    ' Called by
    '   TmrScrollIdle           started by SetpointScrollChanged
    '   chkOutputEnableChanged  when       chkOutputEnable_A_0.CheckedChanged
    '   OpenSystem()            init
    Private Sub ProgramSetpoint(ByVal devnum As Integer)
        Dim errcode As Int32
        Dim NewSetpoint As Double
        Dim NewEnable As Integer
        Dim devtype As DEVTYPE
        Dim type_index As Integer

        If (sysdev_list(devnum).valid <> 0) Then
            devtype = sysdev_list(devnum).dev.devtype
            type_index = sysdev_list(devnum).dev.type_index

            'NewSetpoint = scrollSetpoint_A_0.Value
            'NewEnable = CInt(chkOutputEnable_A_0.Checked)
            NewSetpoint = sysdev_list(devnum).sb_setpoint.Value
            NewEnable = CInt(sysdev_list(devnum).cb_outen.Checked)

            If (sysdev_list(devnum).dev.setpoint_type = SIGTYPE.AMPS) Then
                NewSetpoint = NewSetpoint / AMP2SCROLL_BAR_RANGE ' ((0-max)*10000)/10000 = Amps....scrolbar range = 0 to max*10000
            End If

            'errcode = C3600_SetOutputRaw(sysid, devcfg.devtype, devcfg.type_index, raw, NewSetpoint, NewEnable, INFINITE_WAIT)
            errcode = C3600_SetOutput(sysid, devtype, type_index, NewSetpoint, NewEnable, INFINITE_WAIT)
        End If
    End Sub

    Private Sub TmrScrollIdle_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles TmrScrollIdle.Tick
        TmrScrollIdle.Stop()
        ProgramSetpoint(timer_devnum)
    End Sub

    'Private Sub chkOutputEnableChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles chkOutputEnable_A_0.CheckedChanged, chkOutputEnable_B_0.CheckedChanged, chkOutputEnable_F_0.CheckedChanged, chkOutputEnable_A_1.CheckedChanged, chkOutputEnable_B_1.CheckedChanged, chkOutputEnable_F_1.CheckedChanged
    Private Sub chkOutputEnableChanged(ByVal sender As System.Object, ByVal e As System.EventArgs)
        Dim devnum As Integer
        devnum = FindDevice(sender.name())
        If (init = 0) Then ProgramSetpoint(devnum)
    End Sub

    'Private Function ReadRawSetpointMeter() As Integer
    '    Dim raw As Integer
    '    'C3600_GetMeter(sysid, devcfg.devtype, devcfg.type_index, devcfg.setpoint_type, IntPtr.Zero, raw, INFINITE_WAIT)
    '    'C3600_GetMeter(sysid, devcfg.devtype, devcfg.type_index, SIGTYPE.VOLTS, IntPtr.Zero, raw, INFINITE_WAIT)
    '    'C3600_GetMeter(sysid, devcfg.devtype, devcfg.type_index, SIGTYPE.AMPS , IntPtr.Zero, raw, INFINITE_WAIT)
    '    Return raw
    'End Function

  Public Sub OpenSystem_progress(ByVal phase As Int32, ByVal stepnum As Int32, ByVal steptotal As Int32)
    'MsgBox ("Phase: " & phase & "   step: " & stepnum & "   total: " & steptotal )

    ' phase = 0-2
    If (phase = 0) Then
      ' Progress of com port scan to find all ports in system and what there order should be
      ProgressBar1.Maximum = steptotal
      ProgressBar1.Value = stepnum
    End If
    If (phase = 1) Then
      ' Read each ports configuration
      ProgressBar2.Maximum = steptotal
      ProgressBar2.Value = stepnum
    
      ProgressBar1.Value = ProgressBar1.Maximum
      ProgressBar3.Value = 0
    End If
    If (phase = 2) Then
      ' Read each sub-ports configuration
      ProgressBar3.Maximum = steptotal
      ProgressBar3.Value = stepnum
    End If
    'ProgressBar1.Refresh()
    'ProgressBar2.Refresh()
    'ProgressBar3.Refresh()
    Threading.Thread.Sleep(100)

  End Sub

End Class
