//////////////////////////////////////////////////////////////////////////////////
// Module    : c3600.h                                                          //
// Function  : Application include for CPS 3600 API                             //
// Target OS : OS-independent                                                   //
// Author    : Jim Lamberson                                                    //
// Copyright : (C) 2016 CPS                                                     //
//////////////////////////////////////////////////////////////////////////////////

#ifndef _INC_C3600_H_
#define _INC_C3600_H_

// Constants
#define MAX_SYSTEMS                 4           // Maximum number of systems supported by one host computer.
#define MAX_DEVICES                 16          // Maximum number of devices in one system.
#define MODEL_BUFLEN                8           // Size of model number string buffer (including nul terminator).
#define SN_BUFLEN                   13          // Size of serial number string buffer (including nul terminator).
#define INFINITE_WAIT               (~0)        // Unlimited wait time.

// Error codes returned by API functions
#define C3600_ERR_OK                0
#define C3600_ERR_INVALID_PORT      (-1)        // COM port number sent to HalVerifyOpenCom() was larger than MAX_PORTS=256
#define C3600_ERR_INVALID_ID        (-2)        // Invalid system/cluster/unit/device ID.
// available for future use         (-3)
#define C3600_ERR_INVALID_VALUE     (-4)        // Invalid argument value.
#define C3600_ERR_WRITE_FAIL        (-5)        // Cannot write to port.
#define C3600_ERR_READ_FAIL         (-6)        // Cannot read from port.
#define C3600_ERR_READ_SUM          (-7)        // Checksum error in received packet.
#define C3600_ERR_READ_LEN          (-8)        // Length error in received packet.
#define C3600_ERR_READ_TIMEOUT      (-9)        // Timeout waiting for received packet byte.
#define C3600_ERR_ALREADY_OPEN      (-10)       // Attempted to open a system that is already open.
#define C3600_ERR_CLOSED            (-11)       // System is closed.
#define C3600_ERR_CREATE_LOCK       (-12)       // Can't create mutex.
#define C3600_ERR_LOCK_TIMEOUT      (-13)       // Timed out waiting for device lock.
#define C3600_ERR_DEV_UNAVAILABLE   (-14)       // Failed to open System.
#define C3600_ERR_FUNC_UNSUPPORTED  (-15)       // Function unavailable on this device.
#define C3600_ERR_METER_OVERRANGE   (-16)       // Meter ADC input outside normal range.

#define C3600_ERR_LOCK_ABANDONED     (-17)      // The specified object is a mutex object that was not released by the thread that owned the mutex object before the owning thread terminated. Ownership of the mutex object is granted to the calling thread, and the mutex is set to nonsignaled. If the mutex was protecting persistent state information, you should check it for consistency.
#define C3600_ERR_INVALID_HANDLE     (-18)      // CreateFile() returned INVALID_HANDLE_VALUE
#define C3600_ERR_GETCOMMMODEMSTATUS (-19)      // Failed to read DTR/RTS modem control lines
#define C3600_ERR_SETCOMMSTATE       (-20)      // Failed to set DTR/RTS modem control lines
#define C3600_ERR_LOCKING_NULL_DEV   (-21)      // Tried to lock a device set to NULL
#define C3600_ERR_SELECTUNIT_ON_UND  (-22)      // Tried to call selectunit on unit that is not detected
#define C3600_ERR_PORT_GREATER_THAN_MAXPORT (-23) // COM port number sent to HalOpenCom() was larger than MAX_PORTS=256  (Used to be C3600_ERR_INVALID_PORT)

#define C3600_ERR_INTERNAL          (-99)       // System error (use GetLastError() or equivalent to determine system error code).

// Flags returned by C3600_GetFaultFlags()
#define FAULT_VMETER_ADC            0x40        // Voltage metering ADC malfunction.
#define FAULT_IMETER_ADC            0x20        // Current metering ADC malfunction.
#define FAULT_SETPOINT_DAC          0x08        // Setpoint DAC malfunction.

// Flags returned by C3600_GetStatusFlags()
#define STATUS_RESET                0x80        // Device has been reset.
#define STATUS_OVER_TEMP            0x20        // Overtemperature input is asserted.
#define STATUS_FAULT                0x10        // One or more faults detected; call C3600_GetFaultFlags() for details.
#define STATUS_EEBUSY               0x08        // EEPROM write operation in progress.
#define STATUS_OUTPUT_ENABLED       0x04        // Power supply output is enabled.
#define STATUS_OVER_CURRENT         0x02        // Overcurrent input is asserted.
#define STATUS_OVER_VOLTAGE         0x01        // Overvoltage input is asserted.

// Meter types
#define METER_V                     0           // Volt meter.
#define METER_I                     1           // Current meter.
#define METER_T                     2           // Temperature meter.

// Public types ------------------------------------------

typedef enum DEVTYPE {                  // DEVICE OUTPUT TYPES
    DEVTYPE_UNIMPLEMENTED   = 0,        //   Empty device slot.
    DEVTYPE_UNKNOWN         = 1,        //   Device is present, but its stored identity is invalid.
    DEVTYPE_ACCELERATOR     = 2,        //   Accelerator
    DEVTYPE_BIAS            = 3,        //   Bias
    DEVTYPE_EXTRACTOR       = 4,        //   Extractor
    DEVTYPE_FILAMENT        = 5,        //   Filament
    DEVTYPE_LENS            = 6,        //   Lens
    DEVTYPE_SUPPRESSOR      = 7,        //   Suppressor
    NUM_DEVTYPES
} DEVTYPE;

typedef enum SIGTYPE {                  // SETPOINT TYPE
    VOLTS   = 0,
    AMPS    = 1
} SIGTYPE;

typedef struct OUT_RANGE {                  // DEVICE OUTPUT RANGE
    double          max;                    //  Maximum output.
    double          min;                    //  Minimum output.
} OUT_RANGE;

typedef struct DEV_INFO {                   // DEVICE CONSTANTS
    OUT_RANGE       range[2];               //  Voltage [0] and current [1] output ranges in volts and amps.
    SIGTYPE         sptype;                 //  Setpoint type: VOLTS | AMPS.
    int             is_detected;            //  Device was detected during system open.
    int             is_valid;               //  Device configuration data has valid checksum.
    unsigned short  fwversion;              //  Firmware version number.
    int             devpolarity;            //  Polarity of device. 0=positive, 1=negative.
} DEV_INFO;

typedef struct SYS_INFO {                   // SYSTEM CONSTANTS
    int             is_valid;               //  System configuration data has valid checksum.
    char            model[MODEL_BUFLEN];    //  Power supply model string.
    char            serial_num[SN_BUFLEN];  //  Serial number string.
} SYS_INFO;

// API functions ---------------------------------------

#ifndef APIFUNC
#define APIFUNC
#endif

#ifndef CCONV
#define CCONV
#endif

#ifdef __cplusplus
extern "C" {
#endif

APIFUNC int CCONV C3600_ApiOpen(int *version);
APIFUNC int CCONV C3600_ApiClose(void);
APIFUNC int CCONV C3600_OpenSystem(int sysid, int port, int maxwait);
APIFUNC int CCONV C3600_OpenSystem_with_callback(int sysid, int port, int maxwait, long cbAddress);
APIFUNC int CCONV C3600_CloseSystem(int sysid);
APIFUNC int CCONV C3600_GetSysInfo(int sysid, SYS_INFO *info, int maxwait);
APIFUNC int CCONV C3600_GetDevInfo(int sysid, DEVTYPE devtype, int devindex, DEV_INFO *info, int maxwait);
APIFUNC int CCONV C3600_GetFaultFlags(int sysid, DEVTYPE devtype, int devindex, int *flags, int maxwait);
APIFUNC int CCONV C3600_GetStatusFlags(int sysid, DEVTYPE devtype, int devindex, int *flags, int maxwait);
APIFUNC int CCONV C3600_GetMeter(int sysid, DEVTYPE devtype, int devindex, int meterid, double *data, int *raw, int maxwait);
APIFUNC int CCONV C3600_GetOutput(int sysid, DEVTYPE devtype, int devindex, double *data, int *raw, int *enab, int maxwait);
APIFUNC int CCONV C3600_SetOutput(int sysid, DEVTYPE devtype, int devindex, double data, int enab, int maxwait);
APIFUNC int CCONV C3600_SetOutputRaw(int sysid, DEVTYPE devtype, int devindex, int raw, double *data, int enab, int maxwait);
APIFUNC int CCONV C3600_SetWatchdog(int sysid, int msec, int maxwait);
APIFUNC int CCONV C3600_SysReset(int sysid, int hard, int maxwait);
APIFUNC int CCONV C3600_GetAlarm(int sysid, int *alarm, int maxwait);
APIFUNC const char * CCONV C3600_GetErrorString(int errcode);

APIFUNC int CCONV C3600_HalGetLastError(void);

#ifdef __cplusplus
}
#endif

#endif  // ifndef _INC_C3600_H_
